@@ -1,5 +1,5 @@
-from os.path import join
-import gnomeapplet, gtk, gtk.gdk, gconf, gobject
+from os.path import join, getmtime
+from gi.repository import GObject, Gtk, Gdk, GdkPixbuf, GConf, PanelApplet
 from gettext import gettext as _
 import csv
 import locale
@@ -19,16 +19,16 @@ QUOTES_URL="http://finance.yahoo.com/d/quotes.csv?s=%(s)s&f=snc4l1d1t1c1ohgv&e=.
 QUOTES_CSV_FIELDS=["ticker", "label", "currency", ("trade", float), "date", "time", ("variation", float), ("open", float), ("high", float), ("low", float), ("volume", int)]
 
 # based on http://www.johnstowers.co.nz/blog/index.php/2007/03/12/threading-and-pygtk/
-class _IdleObject(gobject.GObject):
+class _IdleObject(GObject.Object):
 	"""
-	Override gobject.GObject to always emit signals in the main thread
+	Override gobject.Object to always emit signals in the main thread
 	by emmitting on an idle handler
 	"""
 	def __init__(self):
-		gobject.GObject.__init__(self)
+		GObject.Object.__init__(self)
 
 	def emit(self, *args):
-		gobject.idle_add(gobject.GObject.emit,self,*args)
+		GObject.idle_add(GObject.Object.emit,self,*args)
 
 class QuotesRetriever(Thread, _IdleObject):
 	"""
@@ -37,7 +37,7 @@ class QuotesRetriever(Thread, _IdleObject):
 	"""
 	__gsignals__ =  {
 			"completed": (
-				gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
+				GObject.SignalFlags.RUN_LAST, GObject.TYPE_NONE, []),
 			# FIXME: We don't monitor progress, yet ...
 			#"progress": (
 			#	gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [
@@ -51,12 +51,14 @@ class QuotesRetriever(Thread, _IdleObject):
 		self.retrieved = False
 		self.data = []
 		self.currencies = []
+		invest.debug("QuotesRetriever created");
 
 	def run(self):
+		invest.debug("QuotesRetriever started");
 		quotes_url = QUOTES_URL % {"s": self.tickers}
 		try:
 			quotes_file = urlopen(quotes_url, proxies = invest.PROXY)
-			self.data = quotes_file.readlines ()
+			self.data = quotes_file.read ()
 			quotes_file.close ()
 		except Exception, msg:
 			invest.debug("Error while retrieving quotes data (url = %s): %s" % (quotes_url, msg))
@@ -65,31 +67,57 @@ class QuotesRetriever(Thread, _IdleObject):
 		self.emit("completed")
 
 
-class QuoteUpdater(gtk.ListStore):
+class QuoteUpdater(Gtk.ListStore):
 	updated = False
 	last_updated = None
 	quotes_valid = False
 	timeout_id = None
 	SYMBOL, LABEL, CURRENCY, TICKER_ONLY, BALANCE, BALANCE_PCT, VALUE, VARIATION_PCT, PB = range(9)
 	def __init__ (self, change_icon_callback, set_tooltip_callback):
-		gtk.ListStore.__init__ (self, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, bool, float, float, float, float, gtk.gdk.Pixbuf)
+		Gtk.ListStore.__init__ (self, GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_STRING, bool, float, float, float, float, GdkPixbuf.Pixbuf)
 		self.set_update_interval(AUTOREFRESH_TIMEOUT)
 		self.change_icon_callback = change_icon_callback
 		self.set_tooltip_callback = set_tooltip_callback
-		self.set_sort_column_id(1, gtk.SORT_ASCENDING)
-		self.refresh()
+		self.set_sort_column_id(1, Gtk.SortType.ASCENDING)
+		self.load()	# read the last cached quotes file
+		self.refresh()	# download a new quotes file, this may fail if disconnected
 
 		# tell the network manager to notify me when network status changes
 		invest.nm.set_statechange_callback(self.nm_state_changed)
 
+	# loads the cached csv file and its last-modification-time as self.last_updated
+	def load(self):
+		invest.debug("Loading quotes");
+		try:
+			f = open(invest.QUOTES_FILE, 'r')
+			data = f.readlines()
+			f.close()
+
+			self.populate(self.parse_yahoo_csv(csv.reader(data)))
+			self.updated = True
+			self.last_updated = datetime.datetime.fromtimestamp(getmtime(invest.QUOTES_FILE))
+			self.update_tooltip()
+		except Exception, msg:
+			invest.error("Could not load the cached quotes file %s: %s" % (invest.QUOTES_FILE, msg) )
+
+	# stores the csv content on disk so it can be used on next start up
+	def save(self, data):
+		invest.debug("Storing quotes")
+		try:
+			f = open(invest.QUOTES_FILE, 'w')
+			f.write(data)
+			f.close()
+		except Exception, msg:
+			invest.error("Could not save the retrieved quotes file to %s: %s" % (invest.QUOTES_FILE, msg) )
+
 	def set_update_interval(self, interval):
 		if self.timeout_id != None:
 			invest.debug("Canceling refresh timer")
-			gobject.source_remove(self.timeout_id)
+			GObject.source_remove(self.timeout_id)
 			self.timeout_id = None
 		if interval > 0:
 			invest.debug("Setting refresh timer to %s:%02d.%03d" % ( interval / 60000, interval % 60000 / 1000, interval % 1000) )
-			self.timeout_id = gobject.timeout_add(interval, self.refresh)
+			self.timeout_id = GObject.timeout_add(interval, self.refresh)
 
 	def nm_state_changed(self):
 		# when nm is online but we do not have an update timer, create it and refresh
@@ -108,12 +136,16 @@ class QuoteUpdater(gtk.ListStore):
 			return False
 
 		if len(invest.STOCKS) == 0:
+			invest.debug("No stocks configured")
 			return True
 
 		tickers = '+'.join(invest.STOCKS.keys())
+		invest.debug("creating QuotesRetriever")
 		quotes_retriever = QuotesRetriever(tickers)
 		quotes_retriever.connect("completed", self.on_retriever_completed)
+		invest.debug("starting QuotesRetriever")
 		quotes_retriever.start()
+		invest.debug("started QuotesRetriever")
 
 		return True
 
@@ -128,16 +160,14 @@ class QuoteUpdater(gtk.ListStore):
 
 	def on_retriever_completed(self, retriever):
 		if retriever.retrieved == False:
-			tooltip = [_('Invest could not connect to Yahoo! Finance')]
-			if self.last_updated != None:
-				# Translators: %s is an hour (%H:%M)
-				tooltip.append(_('Updated at %s') % self.last_updated.strftime("%H:%M"))
-			self.set_tooltip_callback('\n'.join(tooltip))
+			invest.debug("QuotesRetriever failed");
+			self.update_tooltip(_('Invest could not connect to Yahoo! Finance'))
 		else:
-			self.populate(self.parse_yahoo_csv(csv.reader(retriever.data)))
-			self.updated = True
-			self.last_updated = datetime.datetime.now()
-			self.update_tooltip()
+			invest.debug("QuotesRetriever completed");
+			# cache the retrieved csv file
+			self.save(retriever.data)
+			# load the cache and parse it
+			self.load()
 
 	def on_currency_retriever_completed(self, retriever):
 		if retriever.retrieved == False:
@@ -146,7 +176,7 @@ class QuoteUpdater(gtk.ListStore):
 			self.convert_currencies(self.parse_yahoo_csv(csv.reader(retriever.data)))
 		self.update_tooltip()
 
-	def update_tooltip(self):
+	def update_tooltip(self, msg = None):
 		tooltip = []
 		if self.quotes_count > 0:
 			# Translators: This is share-market jargon. It is the average percentage change of all stock prices. The %s gets replaced with the string value of the change (localized), including the percent sign.
@@ -160,7 +190,10 @@ class QuoteUpdater(gtk.ListStore):
 
 			# Translators: This is share-market jargon. It refers to the total difference between the current price and purchase price for all the shares put together for a particular currency. i.e. How much money would be earned if they were sold right now. The first string is the change value, the second the currency, and the third value is the percentage of the change, formatted using user's locale.
 			tooltip.append(_('Positions balance: %s %s (%s)') % (balance, currency, change))
-		tooltip.append(_('Updated at %s') % self.last_updated.strftime("%H:%M"))
+		if self.last_updated != None:
+			tooltip.append(_('Updated at %s') % self.last_updated.strftime("%H:%M"))
+		if msg != None:
+			tooltip.append(msg)
 		self.set_tooltip_callback('\n'.join(tooltip))
 
 
@@ -238,6 +271,11 @@ class QuoteUpdater(gtk.ListStore):
 			for ticker, val in quote_items:
 				pb = None
 
+				# ignore unknown stocks
+				if ticker not in invest.STOCKS.keys():
+					invest.debug("Observed unknown stock: %s" % ticker)
+					continue
+
 				# get the label of this stock for later reuse
 				label = invest.STOCKS[ticker]["label"]
 				if len(label) == 0:
@@ -275,16 +313,13 @@ class QuoteUpdater(gtk.ListStore):
 						break
 
 				if is_simple_quote:
-					row = self.insert(0, [ticker, label, val["currency"], True, 0, 0, val["trade"], val["variation_pct"], pb])
+					row = self.insert(0, [ticker, label, val["currency"], True, 0.0, 0.0, val["trade"], val["variation_pct"], pb])
 				else:
 					(balance, change) = self.balance(invest.STOCKS[ticker]["purchases"], val["trade"])
 					row = self.insert(0, [ticker, label, val["currency"], False, balance, change, val["trade"], val["variation_pct"], pb])
 					self.add_balance_change(balance, change, val["currency"])
 
-				if len(ticker.split('.')) == 2:
-					url = 'http://ichart.europe.yahoo.com/h?s=%s' % ticker
-				else:
-					url = 'http://ichart.yahoo.com/h?s=%s' % ticker
+				url = 'http://ichart.yahoo.com/h?s=%s' % ticker
 
 				image_retriever = invest.chart.ImageRetriever(url)
 				image_retriever.connect("completed", self.set_pb_callback, row)
@@ -365,15 +400,15 @@ class QuoteUpdater(gtk.ListStore):
 					ticker = self.get_value(iter, self.SYMBOL)
 					value = self.get_value(iter, self.VALUE)
 					(balance, change) = self.balance(invest.STOCKS[ticker]["purchases"], value, rates[currency])
-					self.set(iter, self.BALANCE, balance)
-					self.set(iter, self.BALANCE_PCT, change)
+					self.set_value(iter, self.BALANCE, balance)
+					self.set_value(iter, self.BALANCE_PCT, change)
 					self.add_balance_change(balance, change, target_currency)
 
 				# now, convert the value
 				value = self.get_value(iter, self.VALUE)
 				value *= rates[currency]
-				self.set(iter, self.VALUE, value)
-				self.set(iter, self.CURRENCY, target_currency)
+				self.set_value(iter, self.VALUE, value)
+				self.set_value(iter, self.CURRENCY, target_currency)
 
 			else:
 				# consider non-converted stocks here
@@ -406,6 +441,3 @@ class QuoteUpdater(gtk.ListStore):
 					res = False
 					break
 		return res
-
-if gtk.pygtk_version < (2,8,0):
-	gobject.type_register(QuoteUpdater)
